/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle;

import static java.util.Objects.*;
import static org.apache.commons.lang3.StringUtils.*;

import java.io.IOException;
import java.util.Collection;
import java.util.Optional;

import org.apache.http.HttpHost;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.conn.routing.HttpRoutePlanner;
import org.apache.http.impl.client.BasicAuthCache;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.conn.DefaultProxyRoutePlanner;
import org.apache.http.protocol.HttpContext;

import de.ozgcloud.eingang.formcycle.FormCycleConfirmationResponse;
import de.ozgcloud.eingang.formcycle.FormCycleFormData;
import de.ozgcloud.formcycle.attachment.FormcycleAttachment;
import de.ozgcloud.formcycle.client.OzgCloudResponseHandler;
import de.ozgcloud.formcycle.errorhandling.OzgCloudRequestException;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
public class OzgCloudFormDataHttpClient {

	static final int HTTPCLIENT_TIMEOUT = 10 * 60 * 1000;

	private final OzgCloudConfig config;
	private final SystemPropertiesProvider systemPropertiesProvider;
	private final PortParser portParser;
	private final OzgCloudResponseHandler responseHandler;

	private Optional<ProxyConfig> systemProxyConfig;

	private BasicCredentialsProvider credentialsProvider;
	private HttpClientContext context;
	private HttpHost proxyHost;

	public FormCycleConfirmationResponse send(FormCycleFormData formData, Collection<FormcycleAttachment> attachments,
			Collection<FormcycleAttachment> representations) {

		try (var closeableHttpClient = createCloseableClient()) {
			return closeableHttpClient.execute(buildPostRequest(formData, attachments, representations), responseHandler, getContext());
		} catch (OzgCloudRequestException e) {
			e.setOzgCloudConfig(config);
			throw e;
		} catch (IOException e) {
			throw new OzgCloudRequestException("Cannot send data to formcycle adapter in OZG cloud.", config, e);
		}
	}

	CloseableHttpClient createCloseableClient() {
		var clientBuilder = HttpClientBuilder.create().setDefaultRequestConfig(createHttpClientConfig());
		addProxy(clientBuilder);
		return clientBuilder.build();
	}

	void addProxy(HttpClientBuilder httpClientBuilder) {
		Optional.ofNullable(getProxyConfig()).ifPresent(
				host -> configureProxy(httpClientBuilder));
	}

	void configureProxy(HttpClientBuilder httpClientBuilder) {
		httpClientBuilder.setRoutePlanner(createHttpRouterPlanner());
		httpClientBuilder.setDefaultCredentialsProvider(getCredentialsProvider());
	}

	ProxyConfig getProxyConfig() {
		var proxyConfig = config.getProxyConfig();
		if (nonNull(proxyConfig) && isNoneBlank(proxyConfig.getHost())) {
			return proxyConfig;
		}
		if (isNull(systemProxyConfig)) {
			systemProxyConfig = loadSystemProxyConfig();
		}
		return systemProxyConfig.orElse(null);
	}

	Optional<ProxyConfig> loadSystemProxyConfig() {
		ProxyConfig value = loadHttpsProxyConfig();
		ProxyConfig sysProxyConfig = Optional.ofNullable(value).orElseGet(this::loadHttpProxyConfig);
		return Optional.ofNullable(sysProxyConfig);
	}

	ProxyConfig loadHttpsProxyConfig() {
		var systemProxyHost = systemPropertiesProvider.getHttpsProxyHost();
		if (isBlank(systemProxyHost)) {
			return null;
		}
		return ProxyConfig.builder()
				.host(systemProxyHost)
				.port(portParser.parse(systemPropertiesProvider.getHttpsProxyPort()))
				.user(systemPropertiesProvider.getHttpsProxyUser())
				.password(systemPropertiesProvider.getHttpsProxyPassword())
				.build();
	}

	ProxyConfig loadHttpProxyConfig() {
		var systemProxyHost = systemPropertiesProvider.getHttpProxyHost();
		if (isBlank(systemProxyHost)) {
			return null;
		}
		return ProxyConfig.builder()
				.host(systemProxyHost)
				.port(portParser.parse(systemPropertiesProvider.getHttpProxyPort()))
				.user(systemPropertiesProvider.getHttpProxyUser())
				.password(systemPropertiesProvider.getHttpProxyPassword())
				.build();
	}

	HttpHost getProxyHost() {
		if (isNull(proxyHost)) {
			proxyHost = createProxyHost();
		}
		return proxyHost;
	}

	HttpHost createProxyHost() {
		var proxyConfig = getProxyConfig();
		return new HttpHost(proxyConfig.getHost(), proxyConfig.getPort());
	}

	BasicCredentialsProvider getCredentialsProvider() {
		if (isNull(credentialsProvider)) {
			credentialsProvider = createCredentialsProvider();
		}
		return credentialsProvider;
	}

	BasicCredentialsProvider createCredentialsProvider() {
		var basicCredentialsProvider = new BasicCredentialsProvider();
		var proxyConfig = getProxyConfig();
		basicCredentialsProvider.setCredentials(new AuthScope(getProxyHost()),
				new UsernamePasswordCredentials(proxyConfig.getUser(), proxyConfig.getPassword()));
		return basicCredentialsProvider;
	}

	HttpRoutePlanner createHttpRouterPlanner() {
		return new DefaultProxyRoutePlanner(getProxyHost());
	}

	RequestConfig createHttpClientConfig() {
		return RequestConfig.custom()
				.setConnectTimeout(HTTPCLIENT_TIMEOUT)
				.setConnectionRequestTimeout(HTTPCLIENT_TIMEOUT)
				.setSocketTimeout(HTTPCLIENT_TIMEOUT).build();
	}

	HttpPost buildPostRequest(FormCycleFormData formData, Collection<FormcycleAttachment> attachments,
			Collection<FormcycleAttachment> representations) {
		var builder = new HttpPostRequestBuilder(config).withFormData(formData);
		Optional.ofNullable(attachments).ifPresent(builder::withAttachments);
		Optional.ofNullable(representations).ifPresent(builder::withRepresentations);
		return builder.build();
	}

	HttpContext getContext() {
		if (isNull(context)) {
			context = createContext();
		}
		return context;
	}

	HttpClientContext createContext() {
		var clientContext = HttpClientContext.create();
		if (nonNull(getProxyConfig())) {
			clientContext.setCredentialsProvider(getCredentialsProvider());
			clientContext.setAuthCache(new BasicAuthCache());
		}
		return clientContext;
	}

}
