/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.attachment;

import static de.ozgcloud.formcycle.attachment.FormcycleAttachmentTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.apache.http.entity.ContentType;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.mockito.stubbing.Stubber;

import de.ozgcloud.formcycle.errorhandling.NodeThrewExceptionFactory;
import de.ozgcloud.formcycle.errorhandling.TechnicalException;
import de.xima.fc.entities.Attachment;
import de.xima.fc.entities.AttachmentDatei;
import de.xima.fc.exceptions.NodeThrewException;
import lombok.SneakyThrows;

class AttachmentMapperTest {

	@Spy
	@InjectMocks
	private AttachmentMapper mapper;

	@Mock
	private Attachment attachment;

	@Nested
	class TestMap {

		@Mock
		private AttachmentDatei attachmentDatei;

		@BeforeEach
		void setup() {
			when(attachment.getFileEntity()).thenReturn(attachmentDatei);
		}

		@Test
		void shouldThrowException() {
			assertThrows(TechnicalException.class, () -> mapper.map(attachment));
		}

		@Test
		void shouldReturnResult() {
			var expectedFormcycleAttachment = FormcycleAttachmentTestFactory.create();
			doReturn(expectedFormcycleAttachment).when(mapper).buildAttachment(any());
			when(attachmentDatei.getDaten()).thenReturn(CONTENT);

			var result = mapper.map(attachment);

			assertThat(result).isEqualTo(expectedFormcycleAttachment);
		}
	}

	@Nested
	class TestBuildAttachment {

		@Mock
		private AttachmentDatei attachmentDatei;
		@Mock
		private ContentType contentType;

		@BeforeEach
		void setup() {
			when(attachmentDatei.getDaten()).thenReturn(CONTENT);
			when(attachment.getDateiName()).thenReturn(FILE_NAME);
			when(attachment.getFileEntity()).thenReturn(attachmentDatei);
		}

		@Test
		void shouldCreateAttachmentWithUuid() {
			when(attachment.getUUID()).thenReturn(UUID);

			var ozgAttachment = buildAttachments();

			assertThat(ozgAttachment).hasFieldOrPropertyWithValue(UUID_KEY, UUID);
		}

		@Test
		void shouldCreateAttachmentWithFileName() {
			var ozgAttachment = buildAttachments();

			assertThat(ozgAttachment).hasFieldOrPropertyWithValue(FILE_NAME_KEY, FILE_NAME);
		}

		@Test
		void shouldCreateAttachmentWithData() {
			var ozgAttachment = buildAttachments();

			assertThat(ozgAttachment.getContent()).hasBinaryContent(CONTENT);
		}

		@Test
		void shouldCreateAttachmentWithContentType() {
			doReturn(contentType).when(mapper).getContentType(any());

			var ozgAttachment = buildAttachments();

			assertThat(ozgAttachment.getContentType()).isEqualTo(contentType);
		}

		@SneakyThrows
		private FormcycleAttachment buildAttachments() {
			return mapper.buildAttachment(attachment);
		}
	}

	@Nested
	class TestGetContentType {

		@BeforeEach
		void setup() {
//			when(exceptionFactory.createMissingRepresentationException(any())).thenReturn(mock(NodeThrewException.class));
		}

		@Test
		void shouldThrowException() {
			assertThrows(TechnicalException.class, () -> mapper.getContentType(null));
		}

		@Test
		void shouldReturnResult() {
			var result = mapper.getContentType(FILE_NAME);

			assertThat(result.getMimeType()).isEqualTo(CONTENT_TYPE.getMimeType());
		}
	}
}