package de.ozgcloud.formcycle.formdata;

import static de.ozgcloud.formcycle.formdata.ItemClass.*;
import static de.ozgcloud.formcycle.formdata.StructureMockFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;

import de.xima.fc.form.common.items.XItem;
import lombok.SneakyThrows;

class FormStructureParserTest {

	private static final String NEXT_ITEM_ID = "new-item";

	@Nested
	class TestItems {

		@Test
		void shouldParseInputItem() {
			var expected = FormNodeTestFactory.createBuilder().values(null).build();
			var structureParser = createParser(TEXT_FIELD);

			var formNodes = structureParser.parse();

			assertThat(formNodes).hasSize(1).first().usingRecursiveComparison().isEqualTo(expected);
		}

		@Test
		void shouldParseContainerItem() {
			var expected = FormNodeTestFactory.createBuilder().values(null).build();
			var structureParser = createParser(FIELD_SET);

			var formNodes = structureParser.parse();

			assertThat(formNodes).hasSize(1).first().usingRecursiveComparison().isEqualTo(expected);
		}

		@Test
		void shouldNotParseTextItem() {
			var structureParser = createParser(SPAN);

			var formNodes = structureParser.parse();

			assertThat(formNodes).isEmpty();
		}

		private FormStructureParser createParser(ItemClass itemClass) {
			return new FormStructureParser(createItem(itemClass));
		}
	}

	@Nested
	class TestNestedItems {

		private static final String FS_CONTAINER_ID = "xi-fs-n";
		private static final String NESTED_ITEM_ID1 = "nested1";
		private static final String NESTED_ITEM_ID2 = "nested2";

		private FormStructureParser structureParser;

		@BeforeEach
		void setup() {
			var containerWithItem = initNestedMap();
			structureParser = new FormStructureParser(containerWithItem);
		}

		@Test
		void shouldParseAllInputItems() {
			var structure = new FormStructureParser(StructureMockFactory.create());

			var formNodes = structure.parse();

			assertThat(formNodes).hasSize(2);
		}

		@Test
		void shouldParseNestedFieldSetContainer() {
			var expectedFsNode = FormNodeTestFactory.createBuilder().values(null).itemId(FS_CONTAINER_ID).build();
			var expectedTfNode = FormNodeTestFactory.createBuilder().values(null).itemId(NEXT_ITEM_ID).build();

			var formNodes = structureParser.parse();

			assertThat(formNodes).hasSize(2);
			var containerNodes = formNodes.get(0).getNestedElements();
			assertThat(containerNodes).hasSize(2).first().usingRecursiveComparison().ignoringFields("nestedElements").isEqualTo(expectedFsNode);
			assertThat(containerNodes.get(1)).usingRecursiveComparison().ignoringFields("nestedElements").isEqualTo(expectedTfNode);
		}

		@Test
		void shouldParseItemsInFieldSet() {
			var expectedTfNode = FormNodeTestFactory.createBuilder().values(null).itemId(NESTED_ITEM_ID1).build();

			var fieldSetNodes = structureParser.parse().get(0).getNestedElements().get(0).getNestedElements();

			assertThat(fieldSetNodes).hasSize(1).first().usingRecursiveComparison().ignoringFields("nestedElements").isEqualTo(expectedTfNode);
		}

		private Map<String, XItem> initNestedMap() {
			var containerWithItem = createPageItem();
			containerWithItem.put(CONTAINER_ITEM_ID, getMock(CONTAINER, PAGE_ITEM_ID));
			containerWithItem.put(FS_CONTAINER_ID, getMock(FIELD_SET, FS_CONTAINER_ID, CONTAINER_ITEM_ID, ITEM_TITLE, ITEM_NAME));
			containerWithItem.put(NESTED_ITEM_ID1, getMock(TEXT_FIELD, NESTED_ITEM_ID1, FS_CONTAINER_ID, ITEM_TITLE, ITEM_NAME));
			containerWithItem.put(NESTED_ITEM_ID2, getMock(SPAN, NESTED_ITEM_ID2, FS_CONTAINER_ID, ITEM_TITLE, ITEM_NAME));
			containerWithItem.put(NEXT_ITEM_ID, getMock(TEXT_FIELD, NEXT_ITEM_ID, CONTAINER_ITEM_ID, ITEM_TITLE, ITEM_NAME));
			containerWithItem.put(ITEM_ID, getMock(CHECKBOX, PAGE_ITEM_ID));

			return containerWithItem;
		}

	}

	@Nested
	class TestBuildNode {

		@InjectMocks
		private FormStructureParser structureParser;

		@Test
		void shouldSetId() {
			var formNode = buildNode();

			assertThat(formNode.getItemId()).isEqualTo(ITEM_ID);
		}

		@Test
		void shouldSetTitleInput() {
			var formNode = buildNode();

			assertThat(formNode.getTitle()).isEqualTo(ITEM_TITLE);
		}

		@Test
		void shouldSetTitleFieldSet() {
			var formNode = buildNode(FIELD_SET);

			assertThat(formNode.getTitle()).isEqualTo(ITEM_TITLE);
		}

		@Test
		void shouldSetTitleText() {
			var formNode = buildNode(SPAN);

			assertThat(formNode.getTitle()).isEqualTo(ITEM_TITLE);
		}

		@Test
		void shouldSetName() {
			var formNode = buildNode();

			assertThat(formNode.getName()).isEqualTo(ITEM_NAME);
		}

		@Test
		@DisplayName("should not set title when title property is null")
		void shouldNotSetTitle() {
			var formNode = buildNode(CONTAINER);

			assertThat(formNode.getTitle()).isNull();
		}

		private FormNode buildNode() {
			return buildNode(TEXT_FIELD);
		}

		private FormNode buildNode(ItemClass itemClass) {
			return structureParser.buildNode(getMock(itemClass));
		}
	}

	@Nested
	class TestMethods {

		private static final String PARENT_ID = "parentId";

		@InjectMocks
		private FormStructureParser structureParser;

		private Deque<FormNode> containerNodes;
		private FormNode containerNode;

		@SneakyThrows
		@BeforeEach
		void setup() {
			containerNode = FormNodeTestFactory.create();
			containerNodes = new ArrayDeque<>();
			containerNodes.push(containerNode);
			var stackField = structureParser.getClass().getDeclaredField("containerNodes");
			stackField.setAccessible(true);
			stackField.set(structureParser, containerNodes);
		}

		@Nested
		@DisplayName("GetNextContainerNode")
		class TestNextContainerNode {

			@Test
			@DisplayName("should return empty when container stack is empty")
			void shouldReturnEmpty() {
				injectNodesField();

				var nextContainerNode = structureParser.getNextContainerNode(PARENT_ID);

				assertThat(nextContainerNode).isEmpty();
			}

			@Test
			void shouldReturnParent() {
				var nextContainerNode = structureParser.getNextContainerNode(ITEM_ID);

				assertThat(nextContainerNode).isPresent().get().isEqualTo(containerNode);
			}

			@Test
			void shouldReturnNestedParent() {
				var parentContainer = FormNodeTestFactory.createBuilder().itemId(NEXT_ITEM_ID).build();
				containerNodes.addLast(parentContainer);

				var nextContainerNode = structureParser.getNextContainerNode(NEXT_ITEM_ID);

				assertThat(nextContainerNode).isPresent().get().isEqualTo(parentContainer);
			}

			@Test
			@DisplayName("should return empty when no parent on stack")
			void shouldReturnEmptyNoParents() {
				var parentContainer = FormNodeTestFactory.createBuilder().itemId(NEXT_ITEM_ID).build();
				containerNodes.addLast(parentContainer);

				var nextContainerNode = structureParser.getNextContainerNode("id");

				assertThat(nextContainerNode).isEmpty();
			}

			@Test
			@DisplayName("should clear stack when no parent found")
			void shouldClearStack() {
				structureParser.getNextContainerNode("id");

				assertThat(getStackCollection()).isEmpty();
			}

		}

		@Nested
		@DisplayName("GetNodes")
		class TestGetNodes {

			@Test
			@DisplayName("should return result collection when container stack is empty")
			void shouldReturnResultCollectionStackEmpty() {
				getStackCollection().clear();
				var expectedNodes = injectNodesField();

				var resultNodes = structureParser.getNodes(ITEM_ID);

				assertThat(resultNodes).isEqualTo(expectedNodes);
			}

			@Test
			@DisplayName("should return result collection when no parent found")
			void shouldReturnResultCollectionNoParent() {
				var expectedNodes = injectNodesField();

				var resultNodes = structureParser.getNodes(PARENT_ID);

				assertThat(resultNodes).isEqualTo(expectedNodes);
			}

			@Test
			@DisplayName("should return container collection when parent found")
			void shouldReturnResultContainerCollection() {
				var expectedCollection = containerNode.getNestedElements();

				var resultNodes = structureParser.getNodes(ITEM_ID);

				assertThat(resultNodes).isEqualTo(expectedCollection);
			}
		}

		@SneakyThrows
		private List<FormNode> injectNodesField() {
			List<FormNode> nodes = List.of(FormNodeTestFactory.create());
			var nodesField = structureParser.getClass().getDeclaredField("nodes");
			nodesField.setAccessible(true);
			nodesField.set(structureParser, nodes);

			return nodes;
		}

		@SneakyThrows
		private Deque<FormNode> getStackCollection() {
			var stackField = structureParser.getClass().getDeclaredField("containerNodes");
			stackField.setAccessible(true);
			return (Deque<FormNode>) stackField.get(structureParser);
		}

	}
}